/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ControlTargetBase
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_JointControlTargetBase_H
#define _ARMARX_LIB_RobotAPI_JointControlTargetBase_H

#include <memory>
#include <string>
#include <map>

#include <ArmarXCore/observers/variant/TimedVariant.h>
#include <RobotAPI/interface/units/KinematicUnitInterface.h>

#include "../util.h"
#include "../Constants.h"
#include "../ControlModes.h"

namespace armarx
{
    /**
    * @defgroup Library-RobotUnit RobotUnit
    * @ingroup RobotAPI
    * A description of the library RobotUnit.
    *
    * @class ControlTargetBase
    * @ingroup Library-RobotUnit
    * @brief Brief description of class JointControlTargetBase.
    *
    * Detailed description of class ControlTargetBase.
    */
    class ControlTargetBase
    {
    public:
        virtual void copyTo(ControlTargetBase* target) const = 0;
        virtual std::unique_ptr<ControlTargetBase> clone() const = 0;

        virtual const std::string& getControlMode() const = 0;
        virtual std::string getControlTargetType(bool withoutNamespaceSpecifier = false) const = 0;
        virtual void reset() = 0;
        virtual bool isValid() const = 0;

        /// @brief used to send the data to the DebugObserverTopic and to other Components (e.g. GUI widgets)
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const
        {
            return {};
        }

        template<class T>
        bool isA() const
        {
            return asA<T>();
        }

        template<class T>
        const T* asA() const
        {
            return dynamic_cast<const T*>(this);
        }

        template<class T>
        T* asA()
        {
            return dynamic_cast<T*>(this);
        }

        template<class T, class = typename std::enable_if<std::is_base_of<ControlTargetBase, T>::value>::type>
        void copyTo(std::unique_ptr<T>& target) const
        {
            copyTo(target.get());
        }

        virtual ~ControlTargetBase() = default;
    };

#define DETAIL_ControlTargetBase_DEFAULT_METHOD_IMPLEMENTATION                                      \
    virtual void copyTo(ControlTargetBase* target) const override                                   \
    {                                                                                               \
        using this_type = typename std::decay<decltype(*this)>::type;                               \
        *(target->asA<this_type>()) = *this;                                                        \
    }                                                                                               \
    virtual std::unique_ptr<ControlTargetBase> clone() const override                               \
    {                                                                                               \
        using this_type = typename std::decay<decltype(*this)>::type;                               \
        std::unique_ptr<ControlTargetBase> c {new this_type};                                       \
        ControlTargetBase::copyTo(c);                                                               \
        return c;                                                                                   \
    }                                                                                               \
    virtual std::string getControlTargetType(bool withoutNamespaceSpecifier = false) const override \
    {                                                                                               \
        return getTypeName(*this, withoutNamespaceSpecifier);                                       \
    }

#define make_DummyControlTarget(Suffix,ControlMode)                     \
    class DummyControlTarget##Suffix : public ControlTargetBase         \
    {                                                                   \
    public:                                                             \
        virtual const std::string& getControlMode() const               \
        {                                                               \
            return ControlMode;                                         \
        }                                                               \
        virtual void reset() {}                                         \
        virtual bool isValid() const                                    \
        {                                                               \
            return true;                                                \
        }                                                               \
        DETAIL_ControlTargetBase_DEFAULT_METHOD_IMPLEMENTATION          \
    }
    make_DummyControlTarget(EmergencyStop, ControlModes::EmergencyStop);
    make_DummyControlTarget(StopMovement, ControlModes::StopMovement);
#undef make_DummyControlTarget
}
#endif
