/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ControlDevice.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

namespace armarx
{
    const ControlDevicePtr ControlDevice::NullPtr
    {
        nullptr
    };

    LVL0Controller* ControlDevice::getLVL0EmergencyStopController()
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            lVL0EmergencyStopController,
            "ControlDevice::getLVL0EmergencyStopController called for a ControlDevice without a LVL0Controller with the ControlMode ControlModes::EmergencyStop"
            << " (add a LVL0Controller with this ControlMode)"
        );
        return lVL0EmergencyStopController;
    }

    LVL0Controller* ControlDevice::getLVL0StopMovementController()
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(
            lVL0StopMovementController,
            "ControlDevice::getLVL0StopMovementController called for a ControlDevice without a LVL0Controller with the ControlMode ControlModes::StopMovement"
            << " (add a LVL0Controller with this ControlMode)"
        );
        return lVL0StopMovementController;
    }

    void ControlDevice::rtSetActiveLVL0Controller(LVL0Controller* lvl0)
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(lvl0, "Called ControlDevice::rtSetActiveLVL0Controller with a nullptr (Don't do this)")
        if (activelvl0Controller == lvl0)
        {
            return;
        }
        if (lvl0->parent != this)
        {
            throw std::logic_error
            {
                "ControlDevice(" + getDeviceName() + ")::rtSetActiveLVL0Controller: tried to switch to a lvl0 controller from a different ControlDevice " +
                "(You should only call ControlDevice::rtSetActiveLVL0Controller with LVL0Controllers added to the ControlDevice via ControlDevice::addLVL0Controller)"
            };
        }
        if (activelvl0Controller)
        {
            activelvl0Controller->rtDeactivate();
        }
        lvl0->rtActivate();
        activelvl0Controller = lvl0;
    }

    void ControlDevice::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        auto activeLvl0 = rtGetActiveLVL0Controller();
        ARMARX_CHECK_EXPRESSION(activeLvl0);
        activeLvl0->rtRun(sensorValuesTimestamp, timeSinceLastIteration);
        rtWriteTargetValues(sensorValuesTimestamp, timeSinceLastIteration);
    }

    void ControlDevice::addLVL0Controller(LVL0Controller* lvl0)
    {
        if (!lvl0)
        {
            throw std::invalid_argument
            {
                "ControlDevice(" + getDeviceName() + ")::addLVL0Controller: lvl0 controller is nullptr (Don't try nullptrs as LVL0Controller)"
            };
        }
        if (lvl0->parent)
        {
            throw std::logic_error
            {
                "ControlDevice(" + getDeviceName() + ")::addLVL0Controller: tried to add a lvl0 controller multiple times" +
                "(Don't try to add a LVL0Controller multiple to a ControlDevice)"
            };
        }
        const std::string& mode = lvl0->getControlMode();
        if (lvl0Controllers.has(mode))
        {
            throw std::invalid_argument
            {
                "ControlDevice(" + getDeviceName() + ")::addLVL0Controller: lvl0 controller for mode " + mode + " was already added" +
                "(Don't try to add multiple LVL0Controller with the same ControlMode)"
            };
        }
        lvl0->parent = this;
        lvl0->rtResetTarget();
        if (mode == ControlModes::EmergencyStop)
        {
            lVL0EmergencyStopController = lvl0;
        }
        if (mode == ControlModes::StopMovement)
        {
            lVL0StopMovementController = lvl0;
        }
        if (!lvl0->getControlTarget())
        {
            throw std::logic_error
            {
                "ControlDevice(" + getDeviceName() + ")::addLVL0Controller: The LVL0Controller has no ControlTarget. (mode = " + mode + ")" +
                "(Don't try to add LVL0Controller without a ControlTarget)"
            };
        }
        lvl0Controllers.add(mode, std::move(lvl0));
    }

}
