/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_ControlDevice_H
#define _ARMARX_LIB_RobotAPI_ControlDevice_H

#include <string>
#include <memory>
#include <atomic>

#include"DeviceBase.h"
#include "../util.h"
#include "../ControlModes.h"
#include "../LVL0Controllers/LVL0Controller.h"

namespace armarx
{
    TYPEDEF_PTRS_SHARED(ControlDevice);

    namespace ControlDeviceTags
    {
        using namespace DeviceTags;
        const static std::string PeriodicPosition = "ControlDeviceTag_PeriodicPosition";
        const static std::string CreateNoDefaultController  = "ControlDeviceTag_CreateNoDefaultController";
        const static std::string ForcePositionThroughVelocity  = "ControlDeviceTag_ForcePositionThroughVelocity";
    }

    /**
     * @brief The ControlDevice class represents a logical unit accepting commands via its LVL0Controllers.
     *
     * It holds a set of LVL0Controllers with different Controll modes.
     * It always has a LVL0Controller with mode ControlModes::EmergencyStop which is used in case of an error.
     * It always has a LVL0Controller with mode ControlModes::StopMovement which is used when the controled device should stop any movement.
     *
     * It holds a set of tags.
     * These tags are used by different components to create default controllers, check for abilities or present output to the user.
     */
    class ControlDevice: public virtual DeviceBase
    {
    public:
        /// @brief A static const nullptr in case a const ref to a nullptr needs to be returned
        static const ControlDevicePtr NullPtr;

        /// @brief Create a ControlDevice with the given name
        ControlDevice(const std::string& name): DeviceBase {name} {}
        //controller management
        /// @return the lVL0EmergencyStopController of this ControlDevice
        LVL0Controller* rtGetLVL0EmergencyStopController()
        {
            return lVL0EmergencyStopController;
        }
        /// @return the lVL0EmergencyStopController of this ControlDevice
        LVL0Controller* getLVL0EmergencyStopController();
        /// @return the lVL0StopMovementController of this ControlDevice
        LVL0Controller* rtGetLVL0StopMovementController()
        {
            return lVL0StopMovementController;
        }
        /// @return the lVL0StopMovementController of this ControlDevice
        LVL0Controller* getLVL0StopMovementController();
        /// @return the active LVL0Controller of this ControlDevice
        LVL0Controller* rtGetActiveLVL0Controller()
        {
            return activelvl0Controller;
        }
        /// @return all LVL0Controllers of this ControlDevice
        const std::vector<LVL0Controller*>& rtGetLVL0Controllers()
        {
            return lvl0Controllers.values();
        }
        std::vector<const LVL0Controller*> getLVL0Controllers() const
        {
            return {lvl0Controllers.values().begin(), lvl0Controllers.values().end()};
        }
        /// @return the LVL0Controller for the given mode of this ControlDevice (or nullptr if there is no controller for this mode)
        LVL0Controller* getLVL0Controller(const std::string& mode)
        {
            return lvl0Controllers.at(mode, nullptr);
        }
        const LVL0Controller* getLVL0Controller(const std::string& mode) const
        {
            return lvl0Controllers.at(mode, nullptr);
        }
        LVL0Controller* getLVL0Controller(std::size_t i)
        {
            return lvl0Controllers.at(i);
        }
        const LVL0Controller* getLVL0Controller(std::size_t i) const
        {
            return lvl0Controllers.at(i);
        }
        bool hasLVL0Controller(const std::string& mode) const
        {
            return lvl0Controllers.has(mode);
        }
        const std::vector<std::string>& getControlModes() const
        {
            return lvl0Controllers.keys();
        }
        /**
         * @brief Activates the given LVL0Controller for this device
         * @throw std::logic_error if the lvl0 controller does not belong to this ControlDevice
         */
        virtual void rtSetActiveLVL0Controller(LVL0Controller* lvl0);
        /**
         * @brief runs the active LVL0 Controller and write the target values into the control device
         * @param sensorValuesTimestamp
         * @param timeSinceLastIteration
         * @see writeTargetValues()
         */
        void rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);

        virtual void rtWriteTargetValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) {}
    protected:
        /**
         * @brief adds the lvl0 controller to this ControlDevice
         * @throw std::logic_error if the LVL0Controller was already added to a ControlDevice
         * @throw std::invalid_argument if a LVL0Controller with the same control mode was already added to this device
        */
        void addLVL0Controller(LVL0Controller* lvl0);
    private:
        KeyValueVector<std::string, LVL0Controller*> lvl0Controllers;
        LVL0Controller* activelvl0Controller {nullptr};
        LVL0Controller* lVL0EmergencyStopController {nullptr};
        LVL0Controller* lVL0StopMovementController {nullptr};
    };
}

#endif
