/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_RTThreadTimingsSensorDevice_H
#define _ARMARX_LIB_RobotAPI_RTThreadTimingsSensorDevice_H

#include"SensorDevice.h"
#include "../SensorValues/SensorValueRTThreadTimings.h"

namespace armarx
{
    TYPEDEF_PTRS_SHARED(RTThreadTimingsSensorDevice);

    class RTThreadTimingsSensorDevice: virtual public SensorDevice
    {
    public:
        RTThreadTimingsSensorDevice(const std::string& name): DeviceBase(name), SensorDevice(name) {}
        virtual const SensorValueBase* getSensorValue() const = 0;

        virtual void rtMarkRtLoopStart() = 0;
        virtual void rtMarkRtLoopPreSleep() = 0;

        virtual void rtMarkRtBusSendReceiveStart() = 0;
        virtual void rtMarkRtBusSendReceiveEnd() = 0;
    protected:
        friend class RobotUnit;

        virtual void rtMarkRtSwitchControllerSetupStart() = 0;
        virtual void rtMarkRtSwitchControllerSetupEnd() = 0;

        virtual void rtMarkRtRunLVL1ControllersStart() = 0;
        virtual void rtMarkRtRunLVL1ControllersEnd() = 0;

        virtual void rtMarkRtHandleInvalidTargetsStart() = 0;
        virtual void rtMarkRtHandleInvalidTargetsEnd() = 0;

        virtual void rtMarkRtRunLVL0ControllersStart() = 0;
        virtual void rtMarkRtRunLVL0ControllersEnd() = 0;


        virtual void rtMarkRtReadSensorDeviceValuesStart() = 0;
        virtual void rtMarkRtReadSensorDeviceValuesEnd() = 0;

        virtual void rtMarkRtUpdateSensorAndControlBufferStart() = 0;
        virtual void rtMarkRtUpdateSensorAndControlBufferEnd() = 0;

        virtual void rtMarkRtResetAllTargetsStart() = 0;
        virtual void rtMarkRtResetAllTargetsEnd() = 0;

    };

    template<class SensorValueType = SensorValueRTThreadTimings>
    class RTThreadTimingsSensorDeviceImpl: public RTThreadTimingsSensorDevice
    {
    public:
        static_assert(
            std::is_base_of<SensorValueRTThreadTimings, SensorValueType>::value,
            "The template parameter of RTThreadTimingsSensorDeviceImpl should be derived from SensorValueRTThreadTimings"
        );
        RTThreadTimingsSensorDeviceImpl(const std::string& name): DeviceBase(name), SensorDevice(name), RTThreadTimingsSensorDevice(name) {}
        virtual const SensorValueBase* getSensorValue() const
        {
            return &value;
        }

        SensorValueType value;

        virtual void rtMarkRtLoopStart() override
        {
            std::chrono::microseconds now = MicroNow();
            value.rtLoopRoundTripTime = now - rtLoopStart;
            rtLoopStart = now;
        }
        virtual void rtMarkRtLoopPreSleep() override
        {
            value.rtLoopDuration = MicroNow() - rtLoopStart;
        }
        virtual void rtMarkRtBusSendReceiveStart() override
        {
            std::chrono::microseconds now = MicroNow();
            value.rtBusSendReceiveRoundTripTime = now - rtBusSendReceiveStart;
            rtBusSendReceiveStart = now;
        }
        virtual void rtMarkRtBusSendReceiveEnd() override
        {
            value.rtBusSendReceiveDuration = MicroNow() - rtBusSendReceiveStart;
        }

        static std::chrono::microseconds MicroNow()
        {
            return std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now().time_since_epoch());
        }
    protected:
#define make_markRT_X_Start_End(name)                           \
    virtual void rtMarkRt##name##Start() override                   \
    {                                                           \
        std::chrono::microseconds now = MicroNow();             \
        value.rt##name##RoundTripTime = now - rt##name##Start;  \
        rt##name##Start=now;                                    \
    }                                                           \
    virtual void rtMarkRt##name##End() override                 \
    {                                                           \
        value.rt##name##Duration=MicroNow() - rt##name##Start;  \
    }

        make_markRT_X_Start_End(SwitchControllerSetup)
        make_markRT_X_Start_End(RunLVL1Controllers)
        make_markRT_X_Start_End(HandleInvalidTargets)
        make_markRT_X_Start_End(RunLVL0Controllers)
        make_markRT_X_Start_End(ReadSensorDeviceValues)
        make_markRT_X_Start_End(UpdateSensorAndControlBuffer)
        make_markRT_X_Start_End(ResetAllTargets)
#undef make_markRT_X_Start_End
    protected:
        std::chrono::microseconds rtSwitchControllerSetupStart;
        std::chrono::microseconds rtRunLVL1ControllersStart;
        std::chrono::microseconds rtHandleInvalidTargetsStart;
        std::chrono::microseconds rtRunLVL0ControllersStart;
        std::chrono::microseconds rtBusSendReceiveStart;
        std::chrono::microseconds rtReadSensorDeviceValuesStart;
        std::chrono::microseconds rtUpdateSensorAndControlBufferStart;
        std::chrono::microseconds rtResetAllTargetsStart;
        std::chrono::microseconds rtLoopStart;
    };

}

#endif
