/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::LVL1Controller
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_LVL1Controller_H
#define _ARMARX_LIB_RobotAPI_LVL1Controller_H

#include <map>
#include <atomic>
#include <mutex>
#include <functional>
#include <unordered_set>

#include <boost/optional.hpp>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/util/Registrar.h>
#include <ArmarXCore/core/util/TripleBuffer.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <ArmarXGui/interface/WidgetDescription.h>

#include <RobotAPI/interface/units/RobotUnit/LVL1Controller.h>
#include <RobotAPI/interface/visualization/DebugDrawerInterface.h>
#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

#include "../ControlTargets/ControlTargetBase.h"

//units
#include "../Devices/ControlDevice.h"
#include "../Devices/SensorDevice.h"

#include "../util.h"
#include "../structs.h"

namespace VirtualRobot
{
    class Robot;
}

namespace armarx
{
    TYPEDEF_PTRS_HANDLE(LVL1Controller);
    TYPEDEF_PTRS_HANDLE(LVL1ControllerDescriptionProviderInterface);

    class RobotUnit;

    class LVL1ControllerDescriptionProviderInterface: virtual public Ice::Object
    {
    public:
        virtual ~LVL1ControllerDescriptionProviderInterface() = default;

        virtual ConstSensorDevicePtr   getSensorDevice(const std::string& sensorDeviceName) const = 0;
        virtual const SensorValueBase* getSensorValue(const std::string& sensorDeviceName) const
        {
            auto sd = getSensorDevice(sensorDeviceName);
            return sd ? sd->getSensorValue() : nullptr;
        }
        virtual ConstControlDevicePtr getControlDevice(const std::string& deviceName) = 0;
        virtual ControlTargetBase* getControlTarget(const std::string& deviceName, const std::string& controlMode) = 0;

        virtual std::string getName() const = 0;

        template<class T>
        const T* getSensorValue(const std::string& sensorDeviceName) const
        {
            return dynamic_cast<T*>(getSensorValue(sensorDeviceName));
        }
        template<class T>
        T* getControlTarget(const std::string& deviceName, const std::string& controlMode)
        {
            return dynamic_cast<T*>(getControlTarget(deviceName, controlMode));
        }
    };

    /**
    * @defgroup Library-RobotUnit RobotUnit
    * @ingroup RobotAPI
    * A description of the library RobotUnit.
    */

    /**
    * @ingroup Library-RobotUnit
    * @brief A high level controller writing its results into targes.
    *
    * This is the abstract base class for all LVL1Controllers.
    * It implements basic management routines required by the RobotUnit and some basic ice calls.
    * LVL1Controllers are intantiated and managed by RobotUnits.
    *
    * Each LVL1Controller has to do the following in their constructor:
    * \li Request joints in exactly one control mode from the robot unit (the control mode can't be changed later)
    *     You will receive pointer targets where it can store its results.
    * \li Get pointer to the state of required joints / sensors. (you can't request other data at a later point)
    *     You will receive pointer to the values.
    *
    * Each controller has two parts:
    * \li The RT controll loop
    * \li The NonRT ice communication
    *
    * \section rtcontrollloop The RT controll loop (rtRun)
    * Here you have access to the robots current state and write your results into targets.
    * \warning This part has to satisfy realtime conditions!
    * All realtime able functions of this class have the 'rt' prefix.
    *
    * You must not:
    * \li call any blocking operation
    * \li allocate ram on the heap (since this is a blocking operation)
    * \li resize any datastructure (e.g. vector::resize) (this sometimes allocates ram on the heap)
    * \li insert new datafields into datastructures (e.g. vector::push_back) (this sometimes allocates ram on the heap)
    * \li write to any stream (e.g. ARMARX_VERBOSE, std::cout, print, filestreams) (this sometimes blocks/allocates ram on the heap)
    * \li make network calls (e.g. through ice) (this blocks/allocates ram on the heap) (using begin_... end_... version of ice calls IS NO SOLUTION)
    * \li do any expensive calculations (e.g. calculate IK, run some solver, invert big matrices)
    *
    * \section nonrtpart The NonRT ice communication
    * This part consits of any ice communication.
    * Here you could get new controll parameters or targets from other components.
    *
    * \section rtnrtcomm Communication between RT and NonRT
    * All communication between RT and NonRT has to be lockfree.
    * So you have to use constructs like atomics or TripleBuffers (See armarx::LVL1ControllerWithTripleBuffer).
    *
    * \image html LVL1ControllerGeneralDataFlow.svg "General Dataflow in a LVL1Controller"
    *
    *
    *
    *
    * \image html LVL1ControllerAtomicDataFlow.svg "Dataflow in a LVL1Controller using atomics for communication between RT and NonRT"
    *
    *
    *
    *
    * \image html LVL1ControllerTripleBufferDataFlow.svg "Dataflow in a LVL1Controller using a triple buffer for communication between RT and NonRT"
    *
    * \section expensivework How to do expensive calculations
    * You have to execute expensive calculations in a different worker thread.
    * This thread could calculate target values at its own speed (e.g. 100 Hz).
    *
    * While rtRun runs at a higher frequency (e.g. 1000 Hz) and:
    * \li reads target values
    * \li optionally passes the target values to a PID controller
    * \li writes them to the targets
    * \li sends the sensor values to the worker thread.
    *
    * If you do some additional calculation in rtRun, you maybe need to pass config parameters from NonRT to RT using a nonblocking method.
    *
    * \image html LVL1ControllerWorkerThreadDataFlow.svg "Dataflow in a LVL1Controller using a worker thread"
    *
    */
    class LVL1Controller:
        virtual public LVL1ControllerInterface,
        virtual public ManagedIceObject
    {
    public:
        //add these static functions if you want to provide a gui for construction
        // static WidgetDescription::WidgetPtr GenerateConfigDescription
        // (
        //      const boost::shared_ptr<VirtualRobot::Robot>&,
        //      const std::map<std::string, ConstControlDevicePtr>& controlDevices,
        //      const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
        // )
        // {
        //      return a nullptr / Widget if you want remote configuration but do not need any parameters
        // }
        // static LVL1ControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&)
        // {
        //      return the config that is passed to your controller
        // }
        using GenerateConfigDescriptionFunctionSignature = WidgetDescription::WidgetPtr(*)
                (
                    const boost::shared_ptr<VirtualRobot::Robot>&,
                    const std::map<std::string, ConstControlDevicePtr>& controlDevices,
                    const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
                );
        using GenerateConfigFromVariantsFunctionSignature = LVL1ControllerConfigPtr(*)(const StringVariantBaseMap&);

        static const LVL1ControllerPtr NullPtr;

        virtual ~LVL1Controller() = default;
        //ice interface (must not be called in the rt thread)
        virtual bool isControllerActive(const Ice::Current& = GlobalIceCurrent) const final;
        virtual bool isControllerRequested(const Ice::Current& = GlobalIceCurrent) const final;
        virtual bool isDeletable(const Ice::Current& = GlobalIceCurrent) const final;
        virtual bool hasControllerError(const Ice::Current& = GlobalIceCurrent) const final;

        virtual std::string getClassName(const Ice::Current& = GlobalIceCurrent) const override  = 0;
        virtual std::string getInstanceName(const Ice::Current& = GlobalIceCurrent) const final;

        virtual LVL1ControllerDescription getControllerDescription(const Ice::Current& = GlobalIceCurrent) const final;
        virtual LVL1ControllerStatus getControllerStatus(const Ice::Current& = GlobalIceCurrent) const final;
        virtual LVL1ControllerDescriptionWithStatus getControllerDescriptionWithStatus(const Ice::Current& = GlobalIceCurrent) const final;
        virtual RobotUnitInterfacePrx getRobotUnit(const Ice::Current& = GlobalIceCurrent) const final;

        virtual void activateController(const Ice::Current& = GlobalIceCurrent) const final;
        virtual void deactivateController(const Ice::Current& = GlobalIceCurrent) const final;
        virtual void deleteController(const Ice::Current& = GlobalIceCurrent) const final;

        virtual WidgetDescription::StringWidgetDictionary getFunctionDescriptions(const Ice::Current& = GlobalIceCurrent) const override
        {
            return {};
        }
        virtual void callDescribedFunction(const std::string&, const StringVariantBaseMap&, const Ice::Current& = GlobalIceCurrent) override
        {
        }

        std::size_t getId() const
        {
            return id;
        }

        //c++ interface (local calls) (must be called in the rt thread)
        virtual void rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) = 0;
        virtual void rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
        {
            rtRun(sensorValuesTimestamp, timeSinceLastIteration);
        }
    protected:
        /**
         * @brief This function is called before the controller is activated.
         * You can use it to activate a thread again (DO NOT SPAWN NEW THREADS!) e.g. via a std::atomic_bool.
         */
        virtual void rtPreActivateController() {}
        /**
         * @brief This function is called after the controller is deactivated.
         * You can use it to deactivate a thread (DO NOT JOIN THREADS!) e.g. via a std::atomic_bool.
         */
        virtual void rtPostDeactivateController() {}
    public:
        bool rtUsesControlDevice(std::size_t deviceIndex) const
        {
            return controlDeviceUsedBitmap.at(deviceIndex);
        }

        std::size_t rtGetId() const
        {
            return id;
        }

        //used control devices
        virtual StringStringDictionary getControlDeviceUsedControlModeMap(const Ice::Current& = GlobalIceCurrent) const final;
        const std::vector<char>& getControlDeviceUsedBitmap() const
        {
            return controlDeviceUsedBitmap;
        }
        const std::vector<std::size_t>& rtGetControlDeviceUsedIndices() const
        {
            return controlDeviceUsedIndices;
        }
        const std::vector<std::size_t>& getControlDeviceUsedIndices() const
        {
            return controlDeviceUsedIndices;
        }

        //check for conflict
        boost::optional<std::vector<char>> isNotInConflictWith(const LVL1ControllerPtr& other) const
        {
            return isNotInConflictWith(other->getControlDeviceUsedBitmap());
        }
        boost::optional<std::vector<char>> isNotInConflictWith(const std::vector<char>& used) const;

        template<class ItT>
        static boost::optional<std::vector<char>> AreNotInConflict(ItT first, ItT last);
    protected:
        //publish thread hooks
        virtual void onPublishActivation(const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) {}
        virtual void onPublishDeactivation(const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) {}
        virtual void onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) {}
    private:
        void publish(const SensorAndControl& sac, const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer);
        void deactivatePublish(const DebugDrawerInterfacePrx& draw, const DebugObserverInterfacePrx& observer);

        void rtActivateController();
        void rtDeactivateController();
        void rtDeactivateControllerBecauseOfError();

        //firends
        friend class RobotUnitLVL1ControllerAccess;

        void robotUnitInit(RobotUnit* ru,
                           std::size_t ctrlId,
                           StringStringDictionary ctrlDeviceControlModeMap,
                           std::vector<char> ctrlDeviceUsedBitmap,
                           std::vector<std::size_t> ctrlDeviceUsedIndices,
                           bool ctrlDeletable);

        RobotUnit* robotUnit;
        std::size_t id = std::numeric_limits<std::size_t>::max();
        StringStringDictionary controlDeviceControlModeMap;
        std::vector<char> controlDeviceUsedBitmap;
        std::vector<std::size_t> controlDeviceUsedIndices;

        //this data is filled by the robot unit to provide convenience functions
        std::atomic_bool isActive {false};
        std::atomic_bool isRequested {false};
        std::atomic_bool deactivatedBecauseOfError {false};
        bool deletable {false};

        std::atomic_bool publishActive {false};

        std::atomic_bool statusReportedActive {false};
        std::atomic_bool statusReportedRequested {false};

        // ManagedIceObject interface
    protected:
        std::string getDefaultName() const
        {
            return getClassName();
        }
    public:
        struct IdComp
        {
            template<class PtrT1, class PtrT2>
            bool operator()(PtrT1 l, PtrT2 r) const
            {
                //null is bigger than any other
                if (!l)
                {
                    return false;
                }
                if (!r)
                {
                    return true;
                }
                return l->id < r->id;
            }
        };

        // ManagedIceObject interface
    protected:
        virtual void onInitComponent() override
        {
        }
        virtual void onConnectComponent() override
        {
        }
    };

    class LVL1ControllerRegistryEntry
    {
    private:
        friend class RobotUnit;
        virtual LVL1ControllerPtr create(LVL1ControllerDescriptionProviderInterfacePtr, const LVL1ControllerConfigPtr&, const boost::shared_ptr<VirtualRobot::Robot>&) const = 0;
        virtual WidgetDescription::WidgetPtr GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>&, const std::map<std::string, ConstControlDevicePtr>&, const std::map<std::string, ConstSensorDevicePtr>&) const = 0;
        virtual LVL1ControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&) const = 0;
        virtual bool hasRemoteConfiguration() const = 0;
    };
    using LVL1ControllerRegistry = Registrar<std::unique_ptr<LVL1ControllerRegistryEntry>>;

    template<class ControllerType>
    struct LVL1ControllerRegistration;


    template <typename ControlDataStruct>
    class LVL1ControllerWithTripleBuffer: virtual public LVL1Controller
    {
    public:
        using MutexType = std::recursive_mutex;
        using LockGuardType = std::lock_guard<std::recursive_mutex>;

        LVL1ControllerWithTripleBuffer(const ControlDataStruct& initialCommands = ControlDataStruct()):
            controlDataTripleBuffer {initialCommands}
        {
        }

        virtual void rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override
        {
            rtUpdateControlStruct();
            rtRun(sensorValuesTimestamp, timeSinceLastIteration);
        }

    protected:
        const ControlDataStruct& rtGetControlStruct() const
        {
            return controlDataTripleBuffer.getReadBuffer();
        }
        bool                     rtUpdateControlStruct()
        {
            return controlDataTripleBuffer.updateReadBuffer();
        }

        void writeControlStruct()
        {
            //just lock to be save and reduce the impact of an error
            //also this allows code to unlock the mutex before calling this function
            //(can happen if some lockguard in a scope is used)
            LockGuardType lock(controlDataMutex);
            controlDataTripleBuffer.commitWrite();
        }
        ControlDataStruct& getWriterControlStruct()
        {
            return controlDataTripleBuffer.getWriteBuffer();
        }

        void setControlStruct(const ControlDataStruct& newStruct)
        {
            LockGuardType lock {controlDataMutex};
            getWriterControlStruct() = newStruct;
            writeControlStruct();
        }

        void reinitTripleBuffer(const ControlDataStruct& initial)
        {
            controlDataTripleBuffer.reinitAllBuffers(initial);
        }


        mutable MutexType controlDataMutex;
    private:
        WriteBufferedTripleBuffer<ControlDataStruct> controlDataTripleBuffer;
    };
    template <typename ControlDataStruct>
    using LVL1ControllerWithTripleBufferPtr = IceInternal::Handle<LVL1ControllerWithTripleBuffer<ControlDataStruct>>;
}
#include "LVL1Controller.hpp"
#endif
