/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI
 * @author     Raphael ( raphael dot grimm at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#ifndef _ARMARX_LIB_RobotAPI_SensorValueHolonomicPlatformVelocity_H
#define _ARMARX_LIB_RobotAPI_SensorValueHolonomicPlatformVelocity_H

#include "SensorValueBase.h"

#include <ArmarXCore/core/util/algorithm.h>

namespace armarx
{
    class SensorValueHolonomicPlatformVelocity : virtual public SensorValueBase
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        float velocityX;
        float velocityY;
        float velocityRotation;
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const override
        {
            return
            {
                {"velocityX",        {new TimedVariant(velocityX,timestamp)}},
                {"velocityY", {new TimedVariant(velocityY,timestamp)}},
                {"velocityRotation", {new TimedVariant(velocityRotation,timestamp)}},
            };
        }
    };

    class SensorValueHolonomicPlatformAcceleration : virtual public SensorValueBase
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        float accelerationX;
        float accelerationY;
        float accelerationRotation;
        void deriveAccelerationFromVelocityDelta(float dvx, float dvy, float dvrot, float dt)
        {
            accelerationX = dvx / dt;
            accelerationY = dvy / dt;
            accelerationRotation = dvrot / dt;
        }
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const override
        {
            return
            {
                {"accelerationX", {new TimedVariant(accelerationX,timestamp)}},
                {"accelerationY", {new TimedVariant(accelerationY,timestamp)}},
                {"accelerationRotation", {new TimedVariant(accelerationRotation,timestamp)}},
            };
        }
    };

    class SensorValueHolonomicPlatformRelativePosition : virtual public SensorValueBase
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        float relativePositionX;
        float relativePositionY;
        float relativePositionRotation;
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const override
        {
            return
            {
                {"relativePositionX", {new TimedVariant(relativePositionX,timestamp)}},
                {"relativePositionY", {new TimedVariant(relativePositionY,timestamp)}},
                {"relativePositionRotation", {new TimedVariant(relativePositionRotation,timestamp)}},
            };
        }
    };

    class SensorValueHolonomicPlatformAbsolutePosition : virtual public SensorValueBase
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        float absolutePositionX;
        float absolutePositionY;
        float absolutePositionRotation;
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const override
        {
            return
            {
                {"absolutePositionX", {new TimedVariant(absolutePositionX,timestamp)}},
                {"absolutePositionY", {new TimedVariant(absolutePositionY,timestamp)}},
                {"absolutePositionRotation", {new TimedVariant(absolutePositionRotation,timestamp)}},
            };
        }
    };

    class SensorValueHolonomicPlatform :
        virtual public SensorValueHolonomicPlatformVelocity,
        virtual public SensorValueHolonomicPlatformAcceleration,
        virtual public SensorValueHolonomicPlatformRelativePosition
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION

        void setVelocitiesAndDeriveAcceleration(float vx, float vy, float vrot, float dt)
        {
            deriveAccelerationFromVelocityDelta(vx - velocityX, vy - velocityY, vrot - velocityRotation, dt);
            velocityX = vx;
            velocityY = vy;
            velocityRotation = vrot;
        }
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const override
        {
            auto map = SensorValueHolonomicPlatformVelocity::toVariants(timestamp);
            mergeMaps(map, SensorValueHolonomicPlatformAcceleration::toVariants(timestamp), MergeMapsMode::OverrideNoValues);
            mergeMaps(map, SensorValueHolonomicPlatformRelativePosition::toVariants(timestamp), MergeMapsMode::OverrideNoValues);
            return map;
        }
    };

    class SensorValueHolonomicPlatformWithAbsolutePosition : virtual public SensorValueHolonomicPlatform, virtual public SensorValueHolonomicPlatformAbsolutePosition
    {
    public:
        DETAIL_SensorValueBase_DEFAULT_METHOD_IMPLEMENTATION
        virtual std::map<std::string, VariantBasePtr>   toVariants(const IceUtil::Time& timestamp) const override
        {
            auto map = SensorValueHolonomicPlatform::toVariants(timestamp);
            mergeMaps(map, SensorValueHolonomicPlatformAbsolutePosition::toVariants(timestamp), MergeMapsMode::OverrideNoValues);
            return map;
        }
    };
}
#endif
