/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPlugin
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "RobotUnitControlDevicesWidget.h"
#include "ui_RobotUnitControlDevicesWidget.h"

#include "StyleSheets.h"

#include <QCheckBox>

namespace armarx
{
    RobotUnitControlDevicesWidget::RobotUnitControlDevicesWidget(QWidget* parent) :
        QWidget(parent),
        ui(new Ui::RobotUnitControlDevicesWidget)
    {
        ui->setupUi(this);
        ui->treeWidget->setColumnCount(7);
        QTreeWidgetItem* head = ui->treeWidget->headerItem();
        head->setText(idxName, "Name");
        head->setText(idxTags, "Tags");
        head->setText(idxMode, "Mode");
        head->setText(idxAct, "Act");
        head->setText(idxReq, "Req");
        head->setText(idxType, "Target Taype");
        head->setText(idxVal, "Target Value");

        head->setToolTip(idxName, "The control device's name (green: ok, red: missmatch between requested and active mode, orange: active/requested device not found in the displayed list)");
        head->setToolTip(idxTags, "The control device's tags");
        head->setToolTip(idxMode, "The control device's control mode");
        head->setToolTip(idxAct, "Whether the control mode is active (green: ok, red: missmatch between requested and active mode)");
        head->setToolTip(idxReq, "Whether the control mode is requested (green: ok, red: missmatch between requested and active mode)");
        head->setToolTip(idxType, "The control mode's target type");
        head->setToolTip(idxVal, "The control mode's target value");

        ui->treeWidget->setColumnWidth(idxName, 150);
        ui->treeWidget->setColumnWidth(idxTags, 100);
        ui->treeWidget->setColumnWidth(idxMode, 250);
        ui->treeWidget->setColumnWidth(idxAct, 40);
        ui->treeWidget->setColumnWidth(idxReq, 40);
        ui->treeWidget->setColumnWidth(idxType, 350);
    }

    RobotUnitControlDevicesWidget::~RobotUnitControlDevicesWidget()
    {
        delete ui;
    }

    void RobotUnitControlDevicesWidget::controlDeviceStatusChanged(const ControlDeviceStatus& status)
    {
        std::lock_guard<std::recursive_mutex> guard {statusUpdatesMutex};
        if (statusUpdates[status.deviceName].timestampUSec < status.timestampUSec)
        {
            statusUpdates[status.deviceName] = status;
            QMetaObject::invokeMethod(this, "controlDeviceStatusChanged", Qt::QueuedConnection);
        }
    }

    void RobotUnitControlDevicesWidget::controlDeviceStatusChanged()
    {
        std::lock_guard<std::recursive_mutex> guard {statusUpdatesMutex};
        for (const auto& pair : statusUpdates)
        {
            update(pair.second);
        }
        statusUpdates.clear();
    }

    void RobotUnitControlDevicesWidget::reset(RobotUnitInterfacePrx ru)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        robotUnit = ru;
        //delete all
        ui->treeWidget->clear();
        entries.clear();
        //fill
        if (robotUnit)
        {
            for (ControlDeviceDescription& desc : robotUnit->getControlDeviceDescriptions())
            {
                add(desc);
            }
        }
    }

    void RobotUnitControlDevicesWidget::add(const ControlDeviceDescription& desc)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        if (entries.count(desc.deviceName))
        {
            return;
        }
        entries[desc.deviceName] = new RobotUnitControlDevicesWidgetEntry(*this, *(ui->treeWidget), desc);
    }

    void RobotUnitControlDevicesWidget::update(const ControlDeviceStatus& status)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        if (!robotUnit)
        {
            return;
        }
        if (!entries.count(status.deviceName))
        {
            add(robotUnit->getControlDeviceDescription(status.deviceName));
        }
        entries.at(status.deviceName)->update(status);
    }

    RobotUnitControlDevicesWidgetEntry::RobotUnitControlDevicesWidgetEntry(RobotUnitControlDevicesWidget& parent, QTreeWidget& treeWidget, const ControlDeviceDescription& desc)
        : QObject {&parent}
    {
        header = new QTreeWidgetItem;
        treeWidget.addTopLevelItem(header);
        header->setText(RobotUnitControlDevicesWidget::idxName, QString::fromStdString(desc.deviceName));
        //tags
        {
            QCheckBox* boxTags = new QCheckBox{QString::number(desc.tags.size()) + " Tags"};
            treeWidget.setItemWidget(header, RobotUnitControlDevicesWidget::idxTags, boxTags);
            boxTags->setStyleSheet(checkboxStyleSheet());
            boxTags->setChecked(false);
            connect(boxTags, SIGNAL(clicked(bool)), this, SLOT(hideTagList(bool)));
            for (const auto& tag : desc.tags)
            {
                QTreeWidgetItem* child = new QTreeWidgetItem {{QString::fromStdString(tag)}};
                treeWidget.addTopLevelItem(child);
                tags.emplace_back(child);
            }
        }
        for (const auto& nameToType : desc.contolModeToTargetType)
        {
            ControllerEntry& subEntry = subEntries[nameToType.first];
            subEntry.child = new QTreeWidgetItem;
            header->addChild(subEntry.child);
            subEntry.child->setText(RobotUnitControlDevicesWidget::idxMode, QString::fromStdString(nameToType.first));
            subEntry.child->setText(RobotUnitControlDevicesWidget::idxAct, " ");
            subEntry.child->setText(RobotUnitControlDevicesWidget::idxReq, " ");
            subEntry.child->setText(RobotUnitControlDevicesWidget::idxType, QString::fromStdString(nameToType.second));
            subEntry.value = new VariantWidget;
            treeWidget.setItemWidget(subEntry.child, RobotUnitControlDevicesWidget::idxVal, subEntry.value);
        }
    }

    void RobotUnitControlDevicesWidgetEntry::update(const ControlDeviceStatus& status)
    {
        auto colorNew = (status.activeControlMode == status.requestedControlMode) ? green() : red();
        bool newDevsNotFound = false;
        if (activeMode != status.activeControlMode)
        {
            if (subEntries.count(activeMode))
            {
                subEntries.at(activeMode).child->setText(RobotUnitControlDevicesWidget::idxAct, " ");
                subEntries.at(activeMode).child->setBackgroundColor(RobotUnitControlDevicesWidget::idxAct, transparent());
            }
            activeMode = status.activeControlMode;
            if (subEntries.count(activeMode))
            {
                newDevsNotFound = true;
                subEntries.at(activeMode).child->setText(RobotUnitControlDevicesWidget::idxAct, "X");
            }
        }
        if (requestedMode != status.requestedControlMode)
        {
            if (subEntries.count(requestedMode))
            {
                subEntries.at(requestedMode).child->setText(RobotUnitControlDevicesWidget::idxReq, " ");
                subEntries.at(requestedMode).child->setBackgroundColor(RobotUnitControlDevicesWidget::idxReq, transparent());
            }
            requestedMode = status.requestedControlMode;
            if (subEntries.count(requestedMode))
            {
                newDevsNotFound = true;
                subEntries.at(requestedMode).child->setText(RobotUnitControlDevicesWidget::idxReq, "X");
            }
        }
        //color
        {
            //set the color here, since active may change without requested changen (or the other way) -> above there is no update
            if (subEntries.count(requestedMode))
            {
                subEntries.at(requestedMode).child->setBackgroundColor(RobotUnitControlDevicesWidget::idxReq, colorNew);
            }
            if (subEntries.count(activeMode))
            {
                subEntries.at(activeMode).child->setBackgroundColor(RobotUnitControlDevicesWidget::idxAct, colorNew);
            }
            header->setBackgroundColor(RobotUnitControlDevicesWidget::idxName, newDevsNotFound ? orange() : colorNew);
        }
        for (const auto& pair : status.controlTargetValues)
        {
            if (subEntries.count(pair.first))
            {
                subEntries.at(pair.first).value->update(pair.second);
                subEntries.at(pair.first).value->layout()->setContentsMargins(0, 0, 0, 0);
            }
        }
    }

    bool RobotUnitControlDevicesWidgetEntry::matchName(const QString& name)
    {
        return header->text(RobotUnitControlDevicesWidget::idxName).contains(name, Qt::CaseInsensitive);
    }

    bool RobotUnitControlDevicesWidgetEntry::matchTag(const QString& tag)
    {
        for (const auto& tagit : tags)
        {
            if (tagit->text(RobotUnitControlDevicesWidget::idxTags).contains(tag, Qt::CaseInsensitive))
            {
                return true;
            }
        }
        return false;
    }

    std::set<QTreeWidgetItem*> RobotUnitControlDevicesWidgetEntry::matchMode(const QString& mode)
    {
        std::set<QTreeWidgetItem*> hits;
        for (const auto& pair : subEntries)
        {
            if (pair.second.child->text(RobotUnitControlDevicesWidget::idxMode).contains(mode, Qt::CaseInsensitive))
            {
                hits.emplace(pair.second.child);
            }
        }
        return hits;
    }

    std::set<QTreeWidgetItem*> RobotUnitControlDevicesWidgetEntry::isActiveState(const QString& state)
    {
        std::set<QTreeWidgetItem*> hits;
        for (const auto& pair : subEntries)
        {
            if (pair.second.child->text(RobotUnitControlDevicesWidget::idxAct).contains(state, Qt::CaseInsensitive))
            {
                hits.emplace(pair.second.child);
            }
        }
        return hits;
    }

    std::set<QTreeWidgetItem*> RobotUnitControlDevicesWidgetEntry::isRequestedState(const QString& state)
    {
        std::set<QTreeWidgetItem*> hits;
        for (const auto& pair : subEntries)
        {
            if (pair.second.child->text(RobotUnitControlDevicesWidget::idxReq).contains(state, Qt::CaseInsensitive))
            {
                hits.emplace(pair.second.child);
            }
        }
        return hits;
    }

    std::set<QTreeWidgetItem*> RobotUnitControlDevicesWidgetEntry::matchTargetType(const QString& type)
    {
        std::set<QTreeWidgetItem*> hits;
        for (const auto& pair : subEntries)
        {
            if (pair.second.child->text(RobotUnitControlDevicesWidget::idxType).contains(type, Qt::CaseInsensitive))
            {
                hits.emplace(pair.second.child);
            }
        }
        return hits;
    }

    void RobotUnitControlDevicesWidgetEntry::setVisible(bool vis)
    {
        if (!vis)
        {
            hideTagList(true);
        }
        header->setHidden(!vis);
    }

    void RobotUnitControlDevicesWidgetEntry::setChildVis(bool vis, std::set<QTreeWidgetItem*> children)
    {
        for (const auto& pair : subEntries)
        {
            if (children.count(pair.second.child))
            {
                pair.second.child->setHidden(!vis);
            }
            else
            {
                pair.second.child->setHidden(vis);
            }
        }
    }

    void RobotUnitControlDevicesWidgetEntry::hideTagList(bool hide)
    {
        for (QTreeWidgetItem* tag : tags)
        {
            tag->setHidden(hide);
        }
    }
}
