/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * \package    RobotAPI::gui-plugins::RobotUnitPlugin
 * \author     Raphael ( ufdrv at student dot kit dot edu )
 * \date       2017
 * \copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */
#include "RobotUnitLVL1ControllersWidget.h"
#include "ui_RobotUnitLVL1ControllersWidget.h"
#include "StyleSheets.h"

namespace armarx
{
    RobotUnitLVL1ControllersWidget::RobotUnitLVL1ControllersWidget(QWidget* parent) :
        QWidget(parent),
        ui(new Ui::RobotUnitLVL1ControllersWidget)
    {
        ui->setupUi(this);
        ui->treeWidget->setColumnCount(10);
        QTreeWidgetItem* head = ui->treeWidget->headerItem();

        head->setText(idxName, "Name");
        head->setText(idxClass, "Class");
        head->setText(idxActive, "Act");
        head->setText(idxRequested, "Req");
        head->setText(idxError, "Err");
        head->setText(idxCtrlDev, "Devices");
        head->setText(idxCtrlMode, "Modes");
        head->setText(idxActivate, "");
        head->setText(idxDeactivate, "");
        head->setText(idxDelete, "");

        head->setToolTip(idxName, "The controllers instance name (green: ok, red: active and requested state missmatch or there is an error)");
        head->setToolTip(idxClass, "The controllers class name");
        head->setToolTip(idxActive, "Whether the controller is active (green: ok, red: active and requested state missmatch)");
        head->setToolTip(idxRequested, "Whether the controller is requested (green: ok, red: active and requested state missmatch)");
        head->setToolTip(idxError, "Whether the controller is deactivated due to an error (green: ok, red: there is an error)");
        head->setToolTip(idxCtrlDev, "The control devices used by this controller");
        head->setToolTip(idxCtrlMode, "The controll mode for the control devices used by this controller");
        head->setToolTip(idxActivate, "Button to activate the Controller");
        head->setToolTip(idxDeactivate, "Button to deactivate the Controller");
        head->setToolTip(idxDelete, "Button to delete the Controller");

        ui->treeWidget->setColumnWidth(idxName, 400);
        ui->treeWidget->setColumnWidth(idxClass, 200);
        ui->treeWidget->setColumnWidth(idxCtrlDev, 130);
        ui->treeWidget->setColumnWidth(idxCtrlMode, 150);
        ui->treeWidget->setColumnWidth(idxActivate, 40);
        ui->treeWidget->setColumnWidth(idxDeactivate, 40);
        ui->treeWidget->setColumnWidth(idxDelete, 40);
        ui->treeWidget->setColumnWidth(idxRequested, 40);
        ui->treeWidget->setColumnWidth(idxActive, 40);
        ui->treeWidget->setColumnWidth(idxError, 40);
        // UPGRADE QT5
        //        QHeaderView* headerv = ui->treeWidget->header();
        //        headerv->setSectionResizeMode(idxActivate, QHeaderView::Fixed);
        //        headerv->setSectionResizeMode(idxDeactivate, QHeaderView::Fixed);
        //        headerv->setSectionResizeMode(idxDelete, QHeaderView::Fixed);
        //        headerv->setSectionResizeMode(idxRequested, QHeaderView::Fixed);
        //        headerv->setSectionResizeMode(idxActive, QHeaderView::Fixed);
        //        headerv->setSectionResizeMode(idxError, QHeaderView::Fixed);
    }

    RobotUnitLVL1ControllersWidget::~RobotUnitLVL1ControllersWidget()
    {
        delete ui;
    }

    void RobotUnitLVL1ControllersWidget::lVl1ControllerStatusChanged(const LVL1ControllerStatus& status)
    {
        std::lock_guard<std::recursive_mutex> guard {statusUpdatesMutex};
        if (statusUpdates[status.instanceName].timestampUSec < status.timestampUSec)
        {
            statusUpdates[status.instanceName] = status;
            QMetaObject::invokeMethod(this, "lVl1ControllerStatusChanged", Qt::QueuedConnection);
        }
    }

    void RobotUnitLVL1ControllersWidget::lVl1ControllerStatusChanged()
    {
        std::lock_guard<std::recursive_mutex> guard {statusUpdatesMutex};
        if (!robotUnit)
        {
            return;
        }
        for (const auto& pair : statusUpdates)
        {
            lvl1ControllerCreated(pair.second.instanceName);
            entries.at(pair.second.instanceName)->update(pair.second);
        }
        statusUpdates.clear();
    }

    void RobotUnitLVL1ControllersWidget::lvl1ControllerCreated(std::string name)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        if (!robotUnit)
        {
            return;
        }
        if (!entries.count(name))
        {
            entries[name] = new RobotUnitLVL1ControllersWidgetEntry(*this, *(ui->treeWidget), robotUnit->getLVL1ControllerDescription(name));
            entries.at(name)->update(robotUnit->getLVL1ControllerStatus(name));
        }
    }

    void RobotUnitLVL1ControllersWidget::lvl1ControllerDeleted(std::string name)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        if (entries.count(name))
        {
            entries.at(name)->deleteContent();
            entries.at(name)->deleteLater();
            entries.erase(name);
        }
    }

    void RobotUnitLVL1ControllersWidget::reset(RobotUnitInterfacePrx ru)
    {
        std::lock_guard<std::recursive_mutex> guard {mutex};
        robotUnit = ru;
        //delete all
        ui->treeWidget->clear();
        entries.clear();
        //fill
        if (robotUnit)
        {
            for (const LVL1ControllerDescriptionWithStatus& ds : robotUnit->getLVL1ControllerDescriptionsWithStatuses())
            {
                if (!entries.count(ds.description.instanceName))
                {
                    entries[ds.description.instanceName] = new RobotUnitLVL1ControllersWidgetEntry(*this, *(ui->treeWidget), ds.description);
                    entries.at(ds.description.instanceName)->update(ds.status);
                }
            }
        }
    }

    RobotUnitLVL1ControllersWidgetEntry::RobotUnitLVL1ControllersWidgetEntry(RobotUnitLVL1ControllersWidget& parent, QTreeWidget& treeWidget, const LVL1ControllerDescription& desc)
        :
        QObject(&parent),
        controller {desc.controller}
    {
        ARMARX_CHECK_EXPRESSION(controller);
        header = new QTreeWidgetItem;
        treeWidget.addTopLevelItem(header);
        header->setText(RobotUnitLVL1ControllersWidget::idxName, QString::fromStdString(desc.instanceName));
        header->setText(RobotUnitLVL1ControllersWidget::idxClass, QString::fromStdString(desc.className));
        header->setText(RobotUnitLVL1ControllersWidget::idxActive, "?");
        header->setText(RobotUnitLVL1ControllersWidget::idxRequested, "?");
        header->setText(RobotUnitLVL1ControllersWidget::idxError, "?");

        const std::size_t numDev = desc.controlModeAssignment.size();
        const std::size_t numMod = getMapValues<StringStringDictionary, std::set>(desc.controlModeAssignment).size();
        //device list
        {
            //buttons
            boxDev = new QCheckBox{QString::number(numDev) + " Devices"};
            boxMod = new QCheckBox{QString::number(numMod) + " Modes"};
            treeWidget.setItemWidget(header, RobotUnitLVL1ControllersWidget::idxCtrlDev, boxDev);
            treeWidget.setItemWidget(header, RobotUnitLVL1ControllersWidget::idxCtrlMode, boxMod);
            boxDev->setStyleSheet(checkboxStyleSheet());
            boxMod->setStyleSheet(checkboxStyleSheet());
            boxDev->setChecked(false);
            boxMod->setChecked(false);
            connect(boxDev, SIGNAL(clicked(bool)), this, SLOT(hideDeviceList()));
            connect(boxMod, SIGNAL(clicked(bool)), this, SLOT(hideDeviceList()));
            //list
            devsToModes.reserve(desc.controlModeAssignment.size());
            for (const auto& pair : desc.controlModeAssignment)
            {
                QTreeWidgetItem* child = new QTreeWidgetItem;
                treeWidget.addTopLevelItem(child);
                child->setText(RobotUnitLVL1ControllersWidget::idxCtrlDev, QString::fromStdString(pair.first));
                child->setText(RobotUnitLVL1ControllersWidget::idxCtrlMode, QString::fromStdString(pair.second));
                devsToModes.emplace_back(child);
                child->setHidden(true);
            }
        }
        QPushButton* act = new QPushButton;
        QPushButton* dec = new QPushButton;
        act->setToolTip("Activate the Controller");
        dec->setToolTip("Deactivate the Controller");
        treeWidget.setItemWidget(header, RobotUnitLVL1ControllersWidget::idxActivate, act);
        treeWidget.setItemWidget(header, RobotUnitLVL1ControllersWidget::idxDeactivate, dec);
        act->setIcon(QIcon {QString{":/icons/media-playback-start.ico"}});
        dec->setIcon(QIcon {QString{":/icons/media-playback-pause.ico"}});
        act->setFixedWidth(40);
        dec->setFixedWidth(40);
        act->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
        dec->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
        connect(act, SIGNAL(clicked(bool)), this , SLOT(activateController()));
        connect(dec, SIGNAL(clicked(bool)), this , SLOT(deactivateController()));
        if (desc.deletable)
        {
            QPushButton* del = new QPushButton;
            treeWidget.setItemWidget(header, RobotUnitLVL1ControllersWidget::idxDelete, del);
            del->setToolTip("Delete the Controller");
            del->setIcon(QIcon {QString{":/icons/Trash.svg"}});
            del->setFixedWidth(40);
            del->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
            connect(del, SIGNAL(clicked(bool)), this , SLOT(deleteController()));
        }

        //remote functions
        {
            for (const auto& descr : controller->getFunctionDescriptions())
            {
                QTreeWidgetItem* child = new QTreeWidgetItem;
                header->addChild(child);
                child->setFirstColumnSpanned(true);
                child->setText(0, "Remote function: " + QString::fromStdString(descr.first));

                QTreeWidgetItem* childFnc = new QTreeWidgetItem;
                child->addChild(childFnc);
                childFnc->setFirstColumnSpanned(true);
                treeWidget.setItemWidget(childFnc, 0, new RobotUnitLVL1ControllersWidgetRemoteFunction {descr.first, controller, descr.second});
            }
        }
    }

    void RobotUnitLVL1ControllersWidgetEntry::update(const LVL1ControllerStatus& status)
    {
        header->setText(RobotUnitLVL1ControllersWidget::idxActive, status.active ? "X" : " ");
        header->setText(RobotUnitLVL1ControllersWidget::idxRequested, status.requested ? "X" : " ");
        header->setText(RobotUnitLVL1ControllersWidget::idxError, status.error ? "X" : " ");
        auto acReqColor = (status.active != status.requested) ? red() : green();
        header->setBackground(RobotUnitLVL1ControllersWidget::idxActive, {acReqColor});
        header->setBackground(RobotUnitLVL1ControllersWidget::idxRequested, {acReqColor});
        header->setBackground(RobotUnitLVL1ControllersWidget::idxError, {status.error ? red() : green()});
        //name
        header->setBackground(RobotUnitLVL1ControllersWidget::idxName, {status.error ? red() : acReqColor});
    }

    bool RobotUnitLVL1ControllersWidgetEntry::matchName(const QString& name)
    {
        return header->text(RobotUnitLVL1ControllersWidget::idxName).contains(name, Qt::CaseInsensitive);
    }

    bool RobotUnitLVL1ControllersWidgetEntry::matchClass(const QString& name)
    {
        return header->text(RobotUnitLVL1ControllersWidget::idxClass).contains(name, Qt::CaseInsensitive);
    }

    bool RobotUnitLVL1ControllersWidgetEntry::isActiveState(const QString& state)
    {
        return header->text(RobotUnitLVL1ControllersWidget::idxActive).contains(state, Qt::CaseInsensitive);
    }

    bool RobotUnitLVL1ControllersWidgetEntry::isRequestedState(const QString& state)
    {
        return header->text(RobotUnitLVL1ControllersWidget::idxRequested).contains(state, Qt::CaseInsensitive);
    }

    bool RobotUnitLVL1ControllersWidgetEntry::isErrorState(const QString& state)
    {
        return header->text(RobotUnitLVL1ControllersWidget::idxError).contains(state, Qt::CaseInsensitive);
    }

    bool RobotUnitLVL1ControllersWidgetEntry::matchDevice(const QString& dev)
    {
        for (const auto& elem : devsToModes)
        {
            if (elem->text(RobotUnitLVL1ControllersWidget::idxCtrlDev).contains(dev, Qt::CaseInsensitive))
            {
                return true;
            }
        }
        return false;
    }

    bool RobotUnitLVL1ControllersWidgetEntry::matchMode(const QString& mode)
    {
        for (const auto& elem : devsToModes)
        {
            if (elem->text(RobotUnitLVL1ControllersWidget::idxCtrlMode).contains(mode, Qt::CaseInsensitive))
            {
                return true;
            }
        }
        return false;
    }

    void RobotUnitLVL1ControllersWidgetEntry::deleteContent()
    {
        for (auto item : devsToModes)
        {
            delete item;
        }
        delete header;
    }

    void RobotUnitLVL1ControllersWidgetEntry::setVisible(bool vis)
    {
        setDeviceListVisible(vis);
        header->setHidden(!vis);
    }

    void RobotUnitLVL1ControllersWidgetEntry::activateController()
    {
        ARMARX_CHECK_EXPRESSION(controller);
        controller->activateController();
    }

    void RobotUnitLVL1ControllersWidgetEntry::deactivateController()
    {
        ARMARX_CHECK_EXPRESSION(controller);
        controller->deactivateController();
    }

    void RobotUnitLVL1ControllersWidgetEntry::deleteController()
    {
        ARMARX_CHECK_EXPRESSION(controller);
        controller->deleteController();
    }

    void RobotUnitLVL1ControllersWidgetEntry::hideDeviceList()
    {
        QCheckBox* box = dynamic_cast<QCheckBox*>(sender());
        if (box)
        {
            bool checked = box->isChecked();
            boxDev->setChecked(checked);
            boxMod->setChecked(checked);
            for (QTreeWidgetItem* dev : devsToModes)
            {
                dev->setHidden(!checked);
            }
        }
    }
    void RobotUnitLVL1ControllersWidgetEntry::setDeviceListVisible(bool vis)
    {
        boxDev->setChecked(vis);
        boxMod->setChecked(vis);
        for (QTreeWidgetItem* dev : devsToModes)
        {
            dev->setHidden(!vis);
        }
    }

    RobotUnitLVL1ControllersWidgetRemoteFunction::RobotUnitLVL1ControllersWidgetRemoteFunction(
        const std::string& name,
        const LVL1ControllerInterfacePrx& ctrl,
        const WidgetDescription::WidgetPtr& w
    ):
        functionName {name},
    ctrl {ctrl}
    {
        QHBoxLayout* compress = new QHBoxLayout;
        compress->setContentsMargins(0, 0, 0, 0);
        setLayout(compress);
        //main widget
        {
            QVBoxLayout* l = new QVBoxLayout;
            compress->addLayout(l);
            l->setContentsMargins(0, 0, 0, 0);
            //head layout
            {
                QHBoxLayout* head = new QHBoxLayout;
                head->setContentsMargins(0, 0, 0, 0);
                //exec button
                {
                    QPushButton* execute = new QPushButton;
                    execute->setIcon(QIcon {QString{":/icons/media-playback-start.ico"}});
                    execute->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
                    execute->setFixedWidth(40);
                    execute->setToolTip("Execute this function");
                    connect(execute, SIGNAL(clicked(bool)), this, SLOT(execFunction()));
                    head->addWidget(execute);
                }
                //param change box
                {
                    execOnParamChange = new QCheckBox{"Call on parameter changes"};
                    execOnParamChange->setToolTip("Execute this function whenever a parameter changes");
                    head->addWidget(execOnParamChange);
                }
                head->addItem(new QSpacerItem {0, 0, QSizePolicy::MinimumExpanding, QSizePolicy::Minimum});
                l->addLayout(head);
            }
            params = WidgetDescription::makeDescribedWidget(w, this);
            paramValues = params->getVariants();
            params->setContentsMargins(0, 0, 0, 0);
            l->addWidget(params);
        }
        compress->addItem(new QSpacerItem {0, 0, QSizePolicy::MinimumExpanding, QSizePolicy::Minimum});
    }

    void RobotUnitLVL1ControllersWidgetRemoteFunction::execFunction()
    {
        if (paramValues.empty())
        {
            ARMARX_INFO << deactivateSpam(1) << "calling function " << functionName << " with no parameters\n";
        }
        else
        {
            std::stringstream ss;
            ss  << "calling function " << functionName <<  " with parameters:\n";
            for (const auto& pair : paramValues)
            {
                if (pair.second)
                {

                    if (pair.second->data)
                    {
                        ss << "    '" << pair.first << "' of type " << pair.second->data->ice_id() << "\n";
                    }
                    else
                    {
                        ss << "    '" << pair.first << "' nullptr data \n";
                    }
                }
                else
                {
                    ss << "    '" << pair.first << "' nullptr\n";
                }
            }
            ARMARX_INFO << deactivateSpam(1) << ss.str();
        }
        ctrl->callDescribedFunction(functionName, paramValues);
    }

    void RobotUnitLVL1ControllersWidgetRemoteFunction::valueChangedSlot(std::string name, VariantBasePtr value)
    {
        paramValues[std::move(name)] = std::move(value);
        if (execOnParamChange->isChecked())
        {
            execFunction();
        }
    }
}
