/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_ROBOTAPI_UNITS_RobotUnit_SLICE_
#define _ARMARX_ROBOTAPI_UNITS_RobotUnit_SLICE_

#include <ArmarXCore/interface/core/UserException.ice>
#include <ArmarXCore/interface/core/BasicTypes.ice>

#include <ArmarXGui/interface/WidgetDescription.ice>

#include <RobotAPI/interface/units/RobotUnit/LVL1Controller.ice>

#include <RobotAPI/interface/units/KinematicUnitInterface.ice>
#include <RobotAPI/interface/units/ForceTorqueUnit.ice>
#include <RobotAPI/interface/units/InertialMeasurementUnit.ice>
#include <RobotAPI/interface/units/PlatformUnitInterface.ice>

module armarx
{
    dictionary<string, Ice::StringSeq> ControlDeviceNameToControlModesDictionary;

    dictionary<string, string> ControlDeviceNameToLVL1ControllerNameDictionary;

    struct ControlDeviceDescription
    {
        string deviceName;
        StringStringDictionary contolModeToTargetType;
        Ice::StringSeq tags;
    };
    sequence<ControlDeviceDescription> ControlDeviceDescriptionSeq;
    struct ControlDeviceStatus
    {
        string deviceName;
        string activeControlMode;
        string requestedControlMode;
        StringToStringVariantBaseMapMap controlTargetValues;
        long timestampUSec = 0;
    };
    sequence<ControlDeviceStatus> ControlDeviceStatusSeq;


    struct SensorDeviceDescription
    {
        string deviceName;
        string sensorValueType;
        Ice::StringSeq tags;
    };
    sequence<SensorDeviceDescription> SensorDeviceDescriptionSeq;
    struct SensorDeviceStatus
    {
        string deviceName;
        StringVariantBaseMap sensorValue;
        long timestampUSec = 0;
    };
    sequence<SensorDeviceStatus> SensorDeviceStatusSeq;


    struct LVL1ControllerClassDescription
    {
        string className;
        WidgetDescription::Widget configDescription;
    };
    sequence<LVL1ControllerClassDescription> LVL1ControllerClassDescriptionSeq;

    interface RobotUnitInterface
    {
        //controllers
        //names
        ["cpp:const"] idempotent Ice::StringSeq getLVL1ControllerNames();
        ["cpp:const"] idempotent Ice::StringSeq getRequestedLVL1ControllerNames();
        ["cpp:const"] idempotent Ice::StringSeq getActivatedLVL1ControllerNames();
        //proxy/information
        ["cpp:const"] idempotent LVL1ControllerInterface*               getLVL1Controller(string name);
        ["cpp:const"] idempotent StringLVL1ControllerPrxDictionary      getAllLVL1Controllers();

        ["cpp:const"] idempotent LVL1ControllerStatus                   getLVL1ControllerStatus(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent LVL1ControllerStatusSeq                getLVL1ControllerStatuses();

        ["cpp:const"] idempotent LVL1ControllerDescription              getLVL1ControllerDescription(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent LVL1ControllerDescriptionSeq           getLVL1ControllerDescriptions();

        ["cpp:const"] idempotent LVL1ControllerDescriptionWithStatus    getLVL1ControllerDescriptionWithStatus(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent LVL1ControllerDescriptionWithStatusSeq getLVL1ControllerDescriptionsWithStatuses();

        //devices
        ["cpp:const"] idempotent Ice::StringSeq                         getControlDeviceNames() throws LogicError;
        ["cpp:const"] idempotent ControlDeviceDescription               getControlDeviceDescription(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent ControlDeviceDescriptionSeq            getControlDeviceDescriptions() throws LogicError;
        ["cpp:const"] idempotent ControlDeviceStatus                    getControlDeviceStatus(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent ControlDeviceStatusSeq                 getControlDeviceStatuses() throws LogicError;

        ["cpp:const"] idempotent Ice::StringSeq                         getSensorDeviceNames() throws LogicError;
        ["cpp:const"] idempotent SensorDeviceDescription                getSensorDeviceDescription(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent SensorDeviceDescriptionSeq             getSensorDeviceDescriptions() throws LogicError;
        ["cpp:const"] idempotent SensorDeviceStatus                     getSensorDeviceStatus(string name) throws InvalidArgumentException, LogicError;
        ["cpp:const"] idempotent SensorDeviceStatusSeq                  getSensorDeviceStatuses() throws LogicError;
        //classes
        ["cpp:const"] idempotent Ice::StringSeq                         getLVL1ControllerClassNames();
        ["cpp:const"] idempotent LVL1ControllerClassDescription         getLVL1ControllerClassDescription(string name) throws InvalidArgumentException;
        ["cpp:const"] idempotent LVL1ControllerClassDescriptionSeq      getLVL1ControllerClassDescriptions();

        //switching
        void switchLVL1ControllerSetup(Ice::StringSeq newSetup) throws InvalidArgumentException, LogicError;

        void activateLVL1Controller(string controllerInstanceName) throws InvalidArgumentException, LogicError;
        void activateLVL1Controllers(Ice::StringSeq controllerInstanceNames) throws InvalidArgumentException, LogicError;
        void deactivateLVL1Controller(string controllerInstanceName)throws InvalidArgumentException, LogicError;
        void deactivateLVL1Controllers(Ice::StringSeq controllerInstanceNames)throws InvalidArgumentException, LogicError;

        //creting controllers
        LVL1ControllerInterface* createLVL1Controller(string className, string instanceName, LVL1ControllerConfig config) throws InvalidArgumentException, LogicError;
        LVL1ControllerInterface* createLVL1ControllerFromVariantConfig(string className, string instanceName, StringVariantBaseMap config) throws InvalidArgumentException, LogicError;
        //deleting controllers
        void deleteLVL1Controller(string controllerInstanceName)throws InvalidArgumentException, LogicError;
        void deleteLVL1Controllers(Ice::StringSeq controllerInstanceNames)throws InvalidArgumentException, LogicError;
        //loading
        bool loadLibFromPath(string path);
        bool loadLibFromPackage(string package, string libname);

        //units
        ["cpp:const"] idempotent Object* getUnit(string staticIceId);
        ["cpp:const"] idempotent Ice::ObjectProxySeq getUnits();
        ["cpp:const"] idempotent KinematicUnitInterface* getKinematicUnit();
        ["cpp:const"] idempotent ForceTorqueUnitInterface* getForceTorqueUnit();
        ["cpp:const"] idempotent InertialMeasurementUnitInterface* getInertialMeasurementUnit();
        ["cpp:const"] idempotent PlatformUnitInterface* getPlatformUnit();

        //other
        ["cpp:const"] idempotent string getListenerTopicName();
    };

    interface RobotUnitListener
    {
        void lVl1ControllerStatusChanged(LVL1ControllerStatus status);
        void controlDeviceStatusChanged(ControlDeviceStatus status);
        void sensorDeviceStatusChanged(SensorDeviceStatus status);
        void lvl1ControllerClassAdded(string className);
        void lvl1ControllerCreated(string instanceName);
        void lvl1ControllerDeleted(string instanceName);
    };
};

#endif
